﻿using NLog;
using System;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Threading.Tasks;
using ELTE.ChatRoom.Communication;
using ELTE.ChatRoom.Client.Model;

namespace ELTE.ChatRoom.Client.Model
{
    /// <summary>
    /// Chatszoba megvalósítása TCP kommunikációval.
    /// </summary>
    public class TcpChatRoomModel : IChatRoomModel, IDisposable
    {
        #region Private fields

        private TcpClient _client; // TCP kliens
        private StreamReader _reader; // olvasó
        private StreamWriter _writer; // író
        private String _userName; // felhasználónév
        private CancellationTokenSource _source; // token forrás (az olvasás megállításához)
        private Logger _logger; // eseménynapló
        private Boolean _disposed; // megsemmisítettség jelzése

        #endregion

        #region Constructors

        /// <summary>
        /// Modell példányosítása.
        /// </summary>
        public TcpChatRoomModel()
        {
            _disposed = false;
            IsConnected = false;
            _logger = LogManager.GetLogger("client"); // eseménynapló lekérése
        }

        /// <summary>
        /// Modell megsemmisítése.
        /// </summary>
        ~TcpChatRoomModel()
        {
            Dispose(false);
        }

        #endregion

        #region IChatRoomModel properties

        /// <summary>
        /// Kapcsolat állapotának lekérdezése.
        /// </summary>
        public Boolean IsConnected { get; private set; }

        #endregion

        #region IChatRoomModel events

        /// <summary>
        /// Kapcsolatváltás eseménye.
        /// </summary>
        public event EventHandler ConnectionChanged;

        /// <summary>
        /// Üzenetfogadás eseménye.
        /// </summary>
        public event EventHandler<MessageEventArgs> MessageReceived;

        /// <summary>
        /// Felhasználólista fogadás eseménye.
        /// </summary>
        public event EventHandler<UserListEventArgs> UserListReceived;

        #endregion

        #region IChatRoomModel methods

        /// <summary>
        /// Csatlakozás távoli végponthoz.
        /// </summary>
        /// <param name="address">A távoli gép IP címe.</param>
        /// <param name="port">A távoli gép portszáma.</param>
        public void Connect(String address, Int32 port)
        {
            if (_disposed)
                throw new ObjectDisposedException(GetType().Name);
            if (String.IsNullOrEmpty(address))
                throw new ArgumentNullException("ipAddress", "The address cannot is empty.");

            IPAddress ipAddress;
            if (!IPAddress.TryParse(address, out ipAddress))
                throw new ArgumentException("The IP address is not valid.", "ipAddress");

            if (port < IPEndPoint.MinPort || port > IPEndPoint.MaxPort)
                throw new ArgumentException("The port number is not valid.", "port");

            if (IsConnected) // ha már kapcsolódva van, nem kapcsolódunk újra
                return;

            // létrehozzuk az új kapcsolatot
            try
            {
                _client = new TcpClient();
                _client.Connect(ipAddress, port);

                if (!_client.Connected)
                    return;

                _logger.Info("Connected to server " + address + "."); // információ naplózása

                IsConnected = true;
                OnConnectionChanged(); // jelezzük, hogy kapcsolódtunk

                // példányosítjuk az írót és az olvasót
                NetworkStream stream = _client.GetStream();
                _reader = new StreamReader(stream);
                _writer = new StreamWriter(stream);
                _writer.AutoFlush = true; // automatikus küldés minden írás után

                // elindítjuk az olvasást
                ReadMessages();
            }
            catch (Exception ex)          
            {
                _logger.Error("Exception occured during connection."); // hibaüzenet naplózása
                _logger.Trace(ex, "Exception details");
            }
        }

        /// <summary>
        /// Bejelentkezés.
        /// </summary>
        /// <param name="userName">Felhasználónév.</param>
        public void Login(String userName)
        {
            if (_disposed)
                throw new ObjectDisposedException(GetType().Name);
            if (!IsConnected)
                return;

            _userName = userName;

            Message message = new Message
            {
                 Code = MessageCode.Login,
                 SourceName = _userName
            };

            try
            {
                _writer.WriteLine(Message.Serialize(message));
            }
            catch (Exception ex)
            {
                _logger.Error("Exception occured during login.");
                _logger.Trace(ex, "Exception details");

                OnConnectionChanged();
            }
        }

        /// <summary>
        /// Kijelentkezés.
        /// </summary>
        public void Logout()
        {
            if (_disposed)
                throw new ObjectDisposedException(GetType().Name);
            if (!IsConnected)
                return;

            Message message = new Message
            {
                Code = MessageCode.Logout,
                SourceName = _userName
            };

            try
            {
                _writer.WriteLine(Message.Serialize(message));
            }
            catch (Exception ex)
            {
                _logger.Error("Exception occured during logout.");
                _logger.Trace(ex, "Exception details");

                OnConnectionChanged();
            }
        }

        /// <summary>
        /// Üzenet küldése.
        /// </summary>
        /// <param name="message">Üzenet.</param>
        public void SendMessage(String message)
        {
            SendMessage(null, message);
        }

        /// <summary>
        /// Üzenet küldése.
        /// </summary>
        /// <param name="userName">A felhasználó neve.</param>
        /// <param name="message">Üzenet.</param>
        public void SendMessage(String userName, String message)
        {
            if (_disposed)
                throw new ObjectDisposedException(GetType().Name);
            if (!IsConnected)
                return;

            Message formattedMessage = new Message
            {
                Code = MessageCode.Text,
                SourceName = _userName,
                DestinationName = userName,
                Text = message
            };

            try
            {
                _writer.WriteLine(Message.Serialize(formattedMessage));
            }
            catch (Exception ex)
            {
                _logger.Error("Exception occured during message writing.");
                _logger.Trace(ex, "Exception details");

                OnConnectionChanged();
            }
        }

        /// <summary>
        /// Kapcsolat bontása.
        /// </summary>
        public void Disconnect()
        {
            if (_disposed)
                throw new ObjectDisposedException(GetType().Name);
            if (!IsConnected)
                return;

            _logger.Info("Disconnected from server.");

            if (_source != null)
                _source.Cancel(); // olvasás megállítása

            IsConnected = false;
            OnConnectionChanged();
        }

        /// <summary>
        /// Chatkliens megsemmisítése.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        #endregion

        #region Protected methods

        /// <summary>
        /// Chatkliens megsemmisítése.
        /// </summary>
        protected virtual void Dispose(Boolean disposing)
        {
            if (_disposed)
                return;

            _disposed = true;
            IsConnected = false;

            if (disposing)
            {
                _reader.Dispose();
                _writer.Dispose();
            }

            if (_client != null && _client.Connected)
                _client.Close();
        }

        #endregion

        #region Private methods

        /// <summary>
        /// Üzenetek olvasása.
        /// </summary>
        private void ReadMessages()
        {
            TaskScheduler scheduler = TaskScheduler.FromCurrentSynchronizationContext();
            // a visszaszinkronizáláshoz a jelenlegi szálba

            _source = new CancellationTokenSource();
            CancellationToken token = _source.Token; // a megszakításhoz

            Task.Factory.StartNew(() =>
            {
                try
                {
                    while (!token.IsCancellationRequested) // amíg nem kérnek megszakítást
                    {
                        Message message = Message.Deserialize(_reader.ReadLine());
                        if (message == null)
                            continue;

                        Task.Factory.StartNew(() =>
                        {
                            if (!token.IsCancellationRequested) // ha nem kértek megszakítást
                            {
                                switch (message.Code)
                                {
                                    case MessageCode.Text: // szöveges üzenetet kaptunk
                                        OnMessageReceived(message.SourceName, message.Text, !String.IsNullOrEmpty(message.DestinationName)); // jelezzük, hogy kaptunk üzenetet
                                        break;
                                    case MessageCode.List: // felhasználói listát kaptunk
                                        OnUserListReceived(message.Text);
                                        break;
                                    case MessageCode.Login: // névütközésről kaptunk jelzést
                                        OnMessageReceived(null, message.Text, false);
                                        Disconnect();
                                        break;
                                }
                            }

                        }, token, TaskCreationOptions.None, scheduler);
                        // visszaszinkronizálunk a főszálba
                    }
                }
                catch (Exception ex)
                {
                    _logger.Error("Exception occured during message reading.");
                    _logger.Trace(ex, "Exception details");
                }
            }, token, TaskCreationOptions.LongRunning, TaskScheduler.Default).ContinueWith(_ => OnConnectionChanged(), scheduler);
            // megadjuk a később lefuttatandó tevékenységet
        }

        /// <summary>
        /// Kapcsolatváltás eseménykiváltása.
        /// </summary>
        private void OnConnectionChanged()
        {
            if (ConnectionChanged != null)
                ConnectionChanged(this, EventArgs.Empty);
        }

        /// <summary>
        /// Üzenetfogadás eseménykiváltása.
        /// </summary>
        private void OnMessageReceived(String senderName, String text, Boolean isPrivate)
        {
            if (MessageReceived != null)
                MessageReceived(this, new MessageEventArgs { SenderName = senderName, Text = text, IsPrivate = isPrivate });
        }

        /// <summary>
        /// Felhasználólista fogadása.
        /// </summary>
        private void OnUserListReceived(String text)
        {
            if (UserListReceived != null)
                UserListReceived(this, new UserListEventArgs { Users = text.Split(';') });
        }

        #endregion
    }
}
