﻿using System;
using System.Collections.ObjectModel;
using ELTE.ChatRoom.Client.Model;

namespace ELTE.ChatRoom.Client.ViewModel
{
    /// <summary>
    /// Chatszoba nézetmodell típusa.
    /// </summary>
    public class ChatRoomViewModel : ViewModelBase
    {
        private IChatRoomModel _model;

        /// <summary>
        /// Chatszoba nézetmodell példányosítása.
        /// </summary>
        public ChatRoomViewModel(IChatRoomModel model)
        {
            if (model == null)
                throw new ArgumentNullException(nameof(model));

            _model = model;
            _model.ConnectionChanged += new EventHandler(Model_ConnectionChanged);
            _model.MessageReceived += new EventHandler<MessageEventArgs>(Model_MessageReceived);
            _model.UserListReceived += new EventHandler<UserListEventArgs>(Model_UserListReceived);

            Messages = new ObservableCollection<String>();
            Users = new ObservableCollection<String>();

            SendCommand = new DelegateCommand(param =>
            {
                if (param == null) // ha nincs paraméter, akkor mindenkinek küldjük az üzenetet
                {
                    _model.SendMessage(CurrentText);
                }
                else // ha van paraméter, akkor privát üzenetet küldünk
                {
                    _model.SendMessage(param.ToString(), CurrentText);
                }
            });
            // létrehozzuk a parancsot úgy, hogy a megadott metódusokat használja fel

            ConnectCommand = new DelegateCommand(param =>            
            {
                // egy parancson belül végezzük a kijelentkezést és a bejelentkezést
                if (!_model.IsConnected)
                {
                    _model.Connect(Address, 5000);
                    _model.Login(UserName);
                }
                else
                {
                    _model.Logout();
                    _model.Disconnect();
                }
            });

            CurrentText = "Hello";
            Address = "127.0.0.1";
            UserName = "Anonymus";
            InsertMessage("Welcome to the chatroom! Connect to the server.");
        }

        /// <summary>
        /// Üzeneteket tartalmazó felügyelt gyűjtemény lekérdezése.
        /// </summary>
        public ObservableCollection<String> Messages { get; private set; }

        /// <summary>
        /// Felhasználókat tartalmazó felügyelt gyűjtemény lekérdezése.
        /// </summary>
        public ObservableCollection<String> Users { get; private set; }

        /// <summary>
        /// Felhasználónév lekérdezése, vagy beállítása.
        /// </summary>
        public String UserName { get; set; }

        /// <summary>
        /// Távoli végpont címének lekérdezése, vagy beállítása.
        /// </summary>
        public String Address { get; set; }

        /// <summary>
        /// Aktuális üzenet lekérdezése, vagy beállítása.
        /// </summary>
        public String CurrentText { get; set; }

        /// <summary>
        /// Kapcsolati állapot lekérdezése.
        /// </summary>
        public Boolean IsConnected { get { return _model.IsConnected; } }

        /// <summary>
        /// Kapcsolhatóság lekérdezése.
        /// </summary>
        public Boolean CanConnect { get { return !_model.IsConnected; } }

        /// <summary>
        /// Üzenetküldés parancsának lekérdezése.
        /// </summary>
        public DelegateCommand SendCommand { get ; private set; }

        /// <summary>
        /// Kapcsolódás parancsának lekérdezése.
        /// </summary>
        public DelegateCommand ConnectCommand { get ; private set; }

        /// <summary>
        /// Üzenetfogadás eseménykezelője.
        /// </summary>
        private void Model_MessageReceived(object sender, MessageEventArgs e)
        {
            if (String.IsNullOrEmpty(e.SenderName)) // ha nincs küldő, akkor rendszerüzenet
                InsertMessage(e.Text);
            else if (e.IsPrivate)
                InsertMessage(e.SenderName + "->" + UserName, e.Text, e.IsPrivate);
            else
                InsertMessage(e.SenderName, e.Text, e.IsPrivate);
        }

        /// <summary>
        /// Felhasználólista fogadás eseménykezelője.
        /// </summary>
        private void Model_UserListReceived(object sender, UserListEventArgs e)
        {
            Users.Clear();
            Users = new ObservableCollection<String>(e.Users);
            OnPropertyChanged("Users");
        }

        /// <summary>
        /// Kapcsolatváltás eseménykezelője.
        /// </summary>
        private void Model_ConnectionChanged(object sender, EventArgs e)
        {
            if (!_model.IsConnected)
                InsertMessage("Connection closed.");
            else
                InsertMessage("Connection open.");

            Users.Clear();
            OnPropertyChanged("Users");

            OnPropertyChanged("IsConnected");
            OnPropertyChanged("CanConnect");
        }

        /// <summary>
        /// Üzenet beszúrása.
        /// </summary>
        /// <param name="text">Üzenet.</param>
        private void InsertMessage(String text)
        {
            Messages.Insert(0, "[" + DateTime.Now.ToString("HH:mm:ss") + "]: " + text);
        }

        /// <summary>
        /// Üzenet beszúrása.
        /// </summary>
        /// <param name="userName">Felhasználó.</param>
        /// <param name="text">Üzenet.</param>
        /// <param name="isPrivate">Privát-e.</param>
        private void InsertMessage(String userName, String text, Boolean isPrivate)
        {
            Messages.Insert(0, 
                (isPrivate ? "** " : String.Empty) + // a privát üzenetet két csillaggal jelöljük
                userName + " [" + DateTime.Now.ToString("HH:mm:ss") + "]: " + 
                text);
        }
    }
}
