﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using ELTE.ChatRoom.Communication;

namespace ELTE.ChatRoom.Server
{
    /// <summary>
    /// Chatkliens típusa.
    /// </summary>
    public class ChatClient : IDisposable
    {
        private StreamReader _reader;
        private StreamWriter _writer;
        private TcpClient _client;
        private Boolean _disposed;

        /// <summary>
        /// Felhasználó címének lekérdezése.
        /// </summary>
        public String Address { get; private set; }

        /// <summary>
        /// Felhasználónév lekérdezése, vagy beállítása.
        /// </summary>
        public String UserName { get; private set; }

        /// <summary>
        /// Kapcsolat fennállásának lekérdezése.
        /// </summary>
        public Boolean IsConnected { get { return _client.Connected; } }

        /// <summary>
        /// Bejelentkezettség lekérdezése.
        /// </summary>
        public Boolean IsLoggedIn { get { return !String.IsNullOrEmpty(UserName); } }

        /// <summary>
        /// Chatkliens példányosítása.
        /// </summary>
        /// <param name="client">A TCP kliens.</param>
        public ChatClient(TcpClient client)
        {
            if (client == null)
                throw new ArgumentNullException();

            _client = client;
            NetworkStream stream = client.GetStream();

            // példányosítjuk az írót és az olvasót
            _reader = new StreamReader(stream);
            _writer = new StreamWriter(stream);
            _writer.AutoFlush = true; // automatikus küldés minden írás után
            _disposed = false;

            Address = (_client.Client.RemoteEndPoint as IPEndPoint).Address.ToString();
        }

        /// <summary>
        /// Chatkliens megsemmisítése.
        /// </summary>
        ~ChatClient()
        {
            Dispose(false);
        }

        /// <summary>
        /// Bejelentkezés.
        /// </summary>
        /// <param name="userName">A felhasználónév.</param>
        public void Login(String userName)
        {
            if (_disposed)
                throw new ObjectDisposedException(GetType().Name);
            if (String.IsNullOrEmpty(userName))
                throw new ArgumentException(userName);

            UserName = userName;
        }

        /// <summary>
        /// Üzenet írása.
        /// </summary>
        /// <param name="message">Az üzenet.</param>
        public void Write(Message message)
        {
            if (_disposed)
                throw new ObjectDisposedException(GetType().Name);
            if (message == null)
                throw new ArgumentNullException();

            _writer.WriteLine(Message.Serialize(message));
        }

        /// <summary>
        /// Üzenet olvasása.
        /// </summary>
        public Message Read()
        {
            if (_disposed)
                throw new ObjectDisposedException(GetType().Name);

            return Message.Deserialize(_reader.ReadLine());
        }

        /// <summary>
        /// Chatkliens megsemmisítése.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Chatkliens megsemmisítése.
        /// </summary>
        protected virtual void Dispose(Boolean disposing)
        {
            if (_disposed)
                return;

            _disposed = true;

            _client.Close();

            if (disposing)
            {
                _reader.Dispose();
                _writer.Dispose();
            }
        }
    }
}
