﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading.Tasks;
using ELTE.ChatRoom.Communication;
using NLog;

namespace ELTE.ChatRoom.Server
{
    /// <summary>
    /// Chatszerver típusa. 
    /// </summary>
    public class ChatServer
    {
        private TcpListener _listener;
        private Boolean _isRunning;
        private ConcurrentDictionary<String, ChatClient> _clients;
        private Logger _logger; // eseménynapló

        /// <summary>
        /// Chatszerver példányosítása.
        /// </summary>
        public ChatServer()
        {
            _listener = new TcpListener(IPAddress.Any, 5000);
            _isRunning = false;
            _clients = new ConcurrentDictionary<String, ChatClient>(); // kölcsönös kizárást biztosító gyűjtemény
            _logger = LogManager.GetLogger("server"); // eseménynapló lekérése
        }

        /// <summary>
        /// Szerver indítása.
        /// </summary>
        public void Start()
        {
            _listener.Start();

            Task.Run(() =>
            {
                try
                {
                    _isRunning = true;
                    _logger.Info("Server started."); // logbejegyzés

                    while (_isRunning)
                    {
                        TcpClient tcpClient = _listener.AcceptTcpClient(); // várakozunk a kliensekre
                        ChatClient chatClient = new ChatClient(tcpClient);

                        _logger.Info("Client joined from " + chatClient.Address + "."); // információ naplózása

                        Task.Run(() => HandleClient(chatClient)); // kezeljük a klienst egy új taszkban
                    }
                }
                catch (SocketException ex)
                {
                    _logger.Error("Server listener shut down.");
                    _logger.Trace("Exception occured:", ex);
                }
            });
        }

        /// <summary>
        /// Szerver leállítása.
        /// </summary>
        public void Stop()
        {
            if (_isRunning)
                _listener.Stop();

            _logger.Info("Server stopped.");
            _isRunning = false;
        }

        /// <summary>
        /// Kliens kezelése.
        /// </summary>
        /// <param name="client">Kliens.</param>
        private void HandleClient(ChatClient client)
        {
            try
            {
                while (client.IsConnected) // amíg kapcsolatban van
                {
                    Message message = client.Read(); // kiolvassuk az üzenetet

                    switch (message.Code)
                    {
                        case MessageCode.Login:
                            AddClient(client, message.SourceName);
                            break;

                        case MessageCode.Logout:
                            RemoveClient(client, true);
                            break;

                        case MessageCode.Text:
                            SendMessage(message); // továbbküldjük az üzenetet
                            break;
                    }
                }
            }
            catch // ha megszakadt a kapcsolat
            {
                RemoveClient(client, false);
            }
        }

        /// <summary>
        /// Üzenetküldés.
        /// </summary>
        /// <param name="message">Üzenet.</param>
        private void SendMessage(Message message)
        {
            if (!String.IsNullOrEmpty(message.DestinationName)) // ha van az üzenetnek címzettje, csak neki küldjük el
            {
                try
                {
                    _clients[message.SourceName].Write(message);
                }
                catch
                {
                    RemoveClient(_clients[message.SourceName], false);
                }

                if (message.SourceName != message.DestinationName)
                {
                    try
                    {
                        _clients[message.DestinationName].Write(message);
                    }
                    catch
                    {
                        RemoveClient(_clients[message.DestinationName], false);
                    }
                }
            }
            else // egyébként mindenkinek
            {
                foreach (ChatClient client in _clients.Values)
                {
                    try
                    {
                        client.Write(message);
                    }
                    catch
                    {
                        RemoveClient(client, false);
                    }
                }
            }
        }

        /// <summary>
        /// Felhasználói lista elküldése.
        /// </summary>
        private void SendUserList()
        {
            Message message = new Message
            {
                Code = MessageCode.List,
                SourceName = null,
                DestinationName = null,
                Text = _clients.Count == 0 ? String.Empty : _clients.Keys.Aggregate((x, y) => x + ";" + y) // egy szövegbe tömörítjük a felhasználói listát
            };

            SendMessage(message);
        }

        /// <summary>
        /// Kliens felvétele.
        /// </summary>
        /// <param name="client">A kliens.</param>
        /// <param name="userName">A felhasználónév.</param>
        private void AddClient(ChatClient client, String userName)
        {
            if (!_clients.TryAdd(userName, client)) // ha nem sikerült felvenni a klienst, akkor a név már foglalt
            {
                client.Write(new Message
                {
                    Code = MessageCode.Login,
                    SourceName = userName,
                    DestinationName = null,
                    Text = "User name is already taken."
                }); // erről figyelmeztetjük

                return;
            }

            client.Login(userName); // átvesszük a felhasználónevet

            SendMessage(new Message
            {
                Code = MessageCode.Text,
                SourceName = null,
                DestinationName = null,
                Text = userName + " logged in."
            });
            // jelezzük mindenkinek, hogy bejelentkeztek

            SendUserList(); // kiküldjük a felhasználólistát
        }

        /// <summary>
        /// Kliens törlése.
        /// </summary>
        /// <param name="chatClient">A kliens.</param>
        /// <param name="logout">Kijelentkezés történt-e.</param>
        private void RemoveClient(ChatClient client, Boolean logout)
        {
            _logger.Info("Client disconnected from " + client.Address + ".");

            if (client.IsConnected) // lekapcsoljuk
                client.Dispose();

            if (client.UserName == null || !_clients.ContainsKey(client.UserName))
                return;

            _clients.TryRemove(client.UserName, out client); // töröljük

            // jelzünk mindenki másnak, hogy kilépett
            Message message = new Message
            {
                Code = MessageCode.Text,
                SourceName = null,
                DestinationName = null,
                Text = client.UserName + (logout ? " logged out." : " disconnected.")
            };

            SendMessage(message);

            SendUserList(); // kiküldjük a frissített felhasználólistát
        }
    }
}
